import time

from ..data.config import configurator
from ..data.loader import storages


class CallbackDaemon:

    def __init__(self):
        self.configurator = configurator
        self.cache_time_limit = self.configurator.config['cache_time_limit']
        self.callback_storage = storages.get_storage(self.configurator.config)

    def run_callback_daemon(self):
        while True:
            self.callback_storage.on_callbacks_exceeding_time_limit(self.cache_time_limit)
            time.sleep(3600)



import os
import pathlib
from pathlib import Path

from bestconfig import Config


class CallbackConfigurator:

    def __init__(self):
        self.BASE_DIR = os.path.abspath(os.curdir)
        self.config = Config(*self.config_files).callback_stack

    @property
    def config_files(self):
        return [
            os.path.join(folder, file) for folder in (
                f'{self.BASE_DIR}/data/',
                f'{pathlib.Path(__file__).parent.absolute()}/',
                f'{self.BASE_DIR}/',
            )
            for file in os.listdir(folder) if Path(file).suffix in (
                '.json', '.env', '.ini', '.cfg'
            )
        ]



from .abbreviated_callback_stack import AbbreviatedCallbackStack
from .simple_callback_stack import SimpleCallbackStack
from .callback_stack import CallbackStack


__all__ = ('AbbreviatedCallbackStack', 'SimpleCallbackStack')



from datetime import datetime

from redis import Redis

from .abstract_storage import AbstractCallbackStorage
from ..data.loader import configurator
from ..utils.counter import Counter


class RedisCallbackStorage(AbstractCallbackStorage):

    def __init__(self):
        self._configurator = configurator
        self._config = self._configurator.config
        self._db = self._config.redis_db
        self._redis = Redis(db=self._db)
        self._counter = Counter()
        self._pointers = list()

    def add_callback(self, callback: str):
        if self._check_callback_existing(callback):
            self._redis.hset(callback, mapping=self._prepare_data())

        use_count = self._redis.hget(callback, 'use_count')
        self._increase_use_count(callback, use_count)
        self._redis.expire(callback, self._config.cache_time_limit)

    def remove_callback(self, callback: str):
        if not self._check_callback_existing(callback):
            return
        callback_fields = self._get_callback_fields_on_callback(callback)
        use_count = callback_fields[b'use_count']
        self._reduce_use_count(callback, int(use_count))
        if use_count == 1:
            self._pointers.append(callback_fields[b'abbreviation'].decode('utf-8'))
            self._redis.delete(callback)

    def _reduce_use_count(self, callback: str, use_count: int):
        self._redis.hset(callback, 'use_count', use_count - 1)

    def _increase_use_count(self, callback: str, use_count: int):
        self._redis.hset(callback, 'use_count', use_count + 1)

    def _prepare_data(self) -> dict:
        now = datetime.now()

        while self._check_abbreviation_existing(self._counter.count):
            self._counter.increment()
        return {

            'abbreviation': self._counter.count,
            'created_at': now,
            'updated_ad': now,
            'use_count': 0
        }

    def get_callback(self, abbreviation: str) -> str:
        abbreviation = abbreviation.encode()
        for callback in self._redis.keys():
            if abbreviation == self.get_abbreviation(callback):
                return callback.decode('utf-8')

    def get_abbreviation(self, callback: str) -> str:
        return self._redis.hget(callback, 'abbreviation')

    def _get_callback_fields_on_callback(self, callback: str):
        return self._redis.hgetall(callback)

    def _check_callback_existing(self, callback: str) -> bool:
        return self._redis.exists(callback)

    def _check_abbreviation_existing(self, abbreviation: str) -> bool:
        return bool(self.get_callback(abbreviation))

    def on_callbacks_exceeding_time_limit(self, cache_time_limit):
        self.nullify()
        return cache_time_limit

    def clear(self):
        self._pointers = list()
        self.nullify()
        self._redis.flushall()

    def nullify(self):
        self._counter.nullify()



from datetime import datetime

from .abstract_storage import AbstractCallbackStorage
from ..utils.counter import Counter
from ..utils.db_api.schemas import Callback


class SQLiteCallbackStorage(AbstractCallbackStorage):

    def __init__(self):
        self.table = Callback
        self.table.create_table()
        self.counter = Counter()
        self.pointers = list()

    def add_callback(self, callback: str):
        callback_field = self._get_row_on_callback(callback)
        if not callback_field:
            abbreviation = self.pointers.pop() if len(self.pointers) > 0 else self.counter.count
            callback_field = self.table.create(callback=callback, abbreviation=abbreviation)
            self.counter.increment()
        callback_field.use_counter += 1

    def remove_callback(self, callback: str):
        removed_callback = self._get_row_on_callback(callback)
        if not removed_callback:
            return
        self.pointers.append(removed_callback.abbreviation)
        removed_callback.use_counter -= 1
        if removed_callback.use_counter == 0:
            removed_callback.delete_instance()

    def get_callback(self, abbreviation) -> str:
        return self._get_row_on_abbreviation(abbreviation).callback

    def get_abbreviation(self, callback):
        return self._get_row_on_callback(callback).abbreviation

    def _get_row_on_callback(self, callback: str) -> Callback:
        return self.table.get_or_none(self.table.callback == callback)

    def _get_row_on_abbreviation(self, abbreviation: str) -> Callback:
        return self.table.get_or_none(self.table.abbreviation == abbreviation)

    def on_callbacks_exceeding_time_limit(self, cache_time_limit: int):
        self.table.delete().where(
            datetime.now().timestamp() - self.table.updated_at.to_timestamp() > cache_time_limit
        )

    def clear(self):
        self.table.truncate_table()
        self.pointers = list()
        self.counter.nullify()



import string


class CustomNotation(str):

    symbols = f'{string.digits}{string.ascii_letters}'

    @property
    def base(self):
        return len(self.symbols)

    def convert_to_decimal_notation(self):
        return sum([self.symbols.index(i) * self.base ** j for j, i in enumerate(self[::-1])])

    def convert_from_decimal_notation(self, decimal_number: int):
        x, y = decimal_number // self.base, self.symbols[decimal_number % self.base]
        return y if x == 0 else f'{self.convert_from_decimal_notation(x)}{y}'



from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="aiogram-pagination",
    version="0.1.4",
    packages=find_packages(exclude=('tests', 'tests.*', 'examples.*', 'docs',)),
    author="altcode",
    author_email="cosmosx1328@gmail.com",
    description="Module for deep pagination in an aiogram bot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/altroncode/aiogram-pagination",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries'

    ],
    include_package_data=False,
    install_requires=[
        'aiogram>=2.16',
        'bestconfig>=1.3.5',
    ]
)
